#!/usr/bin/env bash
## Bootstrap WHIZARD and environment
##
## with recommanded packages enabled as default.
##
## Author: Simon Braß <simon.brass@desy.de>
## Last modified: 25th May 2021
##
## CAUTION: We take no garantue for the correct usage or indentation of this script.
## USE AT YOUR OWN RISK! → READ AND UNDERSTAND the following lines.
##
## Requirements: Bash (>5), Unix-System
## Unsupported: Mac OS - I do not care enough to provide such support.
##
## * Description
##
##   Bundle all external dependencies of the Monte Carlo event generator WHIZARD.
##   The bootstrap script downloads, builds and installs all external tools in a consistent way in one (discardable) user place.
##   The script is not meant for global installation, i.e. directly in {/usr,/usr/local}, neither to be run as root or with superuser/sudo permissions.
##   We try our best to veto those cases, but beware of the unconventional user.
##
## * External tools
##
##   - Autotools
##     - m4
##     - Make
##     - Autoconf
##     - Automake
##     - Libtool
##   - OCaml
##   - Root (as mandatory dependency of Delphes or HepMC3)
##   - HepMC2
##   - HepMC3
##   - LHAPDF6
##   - FastJet
##   - Pythia8
##   - Delphes
##   - Rivet (convenience for the author of this script)
##
##   We provide a configuration best suited for *our* tool setting, i.e. WHIZARD on a single computer/laptop.\
##
## * TODO
##
##   - CMake (Build)
##   - Compiler FLAGS (FC, CC, CXX) and CFLAGS, CXXFLAGS, FCFLAGS (at least)
##

## * Environment variables
##
##   - FORCE [False|True] :: Force build of autotools, although the underlying system provides them.
##   - BUILD [Path] :: Path to build directory, defaults to the current working directory.
##   - PREFIX [Path] :: Path to installation directory, defaults to $HOME/local.
##   - NPROC [N] :: Number of parallel processes, defaults to (cores - 1).
##   - ROOT_DIR :: Path to top-level directory of a ROOT installation.
##   - CMAKE_GENERATOR [Ninja|Unix Makefiles] :: Use Ninja per default, fallback to Unix Makefiles.
##   - DEBUG [False|True] :: Developer option.
##

test -z "${FORCE}" && FORCE=False
test -z "${BUILD}" && BUILD="$(pwd)"
test -z "${PREFIX}" && PREFIX="${HOME}/local"
LOG="${BUILD}"/log

test -z "${NPROC}" && NPROC="$(($(nproc) - 1))"

test -n "${DEBUG}" && set -x
if test -z "${CMAKE_GENERATOR}"; then
    command -v ninja 1>/dev/null 2>&1
    if test $? -eq 0; then
        CMAKE_GENERATOR="Ninja"
    else
        CMAKE_GENERATOR="Unix Makefiles"
    fi
fi

# ROOT_DIR must point to a directory containing cmake/RootConfig.cmake
test -z "${ROOT_DIR}" && ROOT_DIR="${PREFIX}"

export PATH="${PREFIX}/bin:${PATH}"
export LD_LIBRARY_PATH="${PREFIX}/lib:${PREFIX}/OpenLoops/lib:${LD_LIBRARY_PATH}"

# #######################################################################
# The locations of the sofware packages
# #######################################################################

# 1. Verify packages (checksum, GPG signature,...)
# 2. Switch to secure HTTP

# Autotools
url_make="http://ftp.gnu.org/gnu/make/make-4.3.tar.gz"
url_m4="http://ftp.gnu.org/gnu/m4/m4-1.4.18.tar.gz"
url_autoconf="http://ftp.gnu.org/gnu/autoconf/autoconf-2.69.tar.gz"
url_automake="http://ftp.gnu.org/gnu/automake/automake-1.16.2.tar.gz"
url_libtool="http://ftp.gnu.org/gnu/libtool/libtool-2.4.6.tar.gz"

# CMake
# url_cmake="https://github.com/Kitware/CMake/releases/download/v3.17.3/cmake-3.17.3.tar.gz"

# O'Caml
url_ocaml="https://github.com/ocaml/ocaml/archive/4.12.0.tar.gz"
dir_ocaml="ocaml-4.12.0"

# External tool dependencies
url_root="https://github.com/root-project/root/archive/v6-24-00.tar.gz"

# Whizard External Dependencies
url_hepmc="https://hepmc.web.cern.ch/hepmc/releases/hepmc2.06.10.tgz"
dir_hepmc="HepMC-2.06.10" # The HepMC packages comes with an enstranged nomenclature.
url_hepmc3="https://hepmc.web.cern.ch/hepmc/releases/HepMC3-3.2.3.tar.gz"
url_lhapdf="https://lhapdf.hepforge.org/downloads/?f=LHAPDF-6.3.0.tar.gz"
url_fastjet="http://fastjet.fr/repo/fastjet-3.3.4.tar.gz"
url_openloops="https://openloops.hepforge.org/downloads?f=OpenLoops-2.1.2.tar.gz"
url_pythia8="http://home.thep.lu.se/~torbjorn/pythia8/pythia8305.tgz"
url_delphes="https://github.com/delphes/delphes/archive/3.4.3pre12.tar.gz"

# WHIZARD
url_whizard="https://launchpad.net/whizard/3.0.x/3.0.0/+download/whizard-3.0.0.tar.gz"

# Analysis Tools
url_rivet="https://gitlab.com/hepcedar/rivetbootstrap/raw/3.1.4/rivet-bootstrap"

# #######################################################################

function fetch_pkg_filename () {
    local url="${1}"
    # --spider: test whether file exists (do not download)
    # -S: print response header
    # We expect a content-disposition header with the field 'filename'.
    filename="$(wget -S --spider ${url} 2>&1 | grep -e 'filename' | cut -d'=' -f2)"
    if test -z "${filename}"; then
        basename "${url}"
    else
        echo "${filename}"
    fi
}

# The actual basename of package is quite cumbersome:
# 1. HepForge uses downloads by HTTP requests, i.e. the filename does not match the basename of the URL,
# 2. Most packages follow along the standard line: naming their archives in them same way as they do for the archived directory.
#    However, there are some deviant packages and they do other things.
#    We have to cover them.
function basename_pkg () {
    local url="${1}"
    local suffix="${2:-.tar.gz}"
    # We except Hepforge and Github URLs where we can spider the header with a corresponding filename field.
    if grep -q "hepforge" <<< "${url}" || grep -q "github" <<< "${url}"; then
        fetch_pkg_filename "${url}" | xargs basename -s "${suffix}"
    else
        basename "${url}" "${suffix}"
    fi
}

function download_untar_pkg () {
    local url="${1}"
    # --content-disposition: Use filename from HTTPS header, if available.
    wget -q --content-disposition "${url}" -O- | tar xz
}

function make_make_install () {
    local config="${1}"
    # Do not place variable config in quotes as it needs expansion.
    ./configure --prefix="${PREFIX}" ${config} && \
        make -j"${NPROC}" -s V=0 && \
        make install -s V=0
}

function cmake_make_install () {
    local config="${1}"
    local cmake_path="${2}"
    # Do not place variable config in quotes as it needs expansion.
    cmake ${config} -G "${CMAKE_GENERATOR}" "${cmake_path}" || return
    case "${CMAKE_GENERATOR}" in
        "Unix Makefiles")
            make -j"${NPROC}" -s V=0 &&
                make install -s V=0
            ;;
        "Ninja")
            ninja -j"${NPROC}" &&
                ninja -j"${NPROC}" install
            ;;
        *)
            echo "Unknown CMAKE generator type: ${CMAKE_GENERATOR} [Ninja, Makefile]."
            exit 1
            ;;
    esac
}

# $1 : Download-URL
# $2 : Configure options
build_generic () {
    local url="${1}"
    local config="${2:-}"
    local name="$(basename_pkg ${url} .tar.gz)"
    test -d "${name}" || download_untar_pkg "${url}"
    pushd "${name}" && \
        make_make_install "${config}" && \
        popd || return
}

# $1 : Download-URL
# $2 : Configure options
build_cmake_generic () {
    local url="${1}"
    local config="${2:-}"
    local name="$(basename_pkg ${url} .tar.gz)"
    test -d "${name}" || download_untar_pkg "${url}"
    mkdir -p "${name}-build" &&
        pushd "${name}-build" && \
        cmake_make_install "${config}" "../${name}" && \
        popd || return
}

build_hepmc2 () {
    local url="${1}"
    local config="${2:-}"
    local name="${dir_hepmc}"
    test -d "${name}" || download_untar_pkg "${url}"
    mkdir -p "${name}-build" && \
        pushd "${name}-build" && \
        cmake_make_install "${config}" "../${name}"
        popd || return
}

build_ocaml () {
    local url="${1}"
    local name="${dir_ocaml}"
    test -d "${name}" || download_untar_pkg "${url}"
    pushd "${name}" && \
        ./configure -prefix ${PREFIX} && \
        make world.opt && \
        make install && \
        popd
}

build_openloops () {
    local url="${1}"
    local name="$(basename_pkg ${url} .tar.gz)"
    local dirname="OpenLoops"
    test -d "${PREFIX}" || mkdir -p "${PREFIX}"
    pushd "${PREFIX}" && \
        test -d "${dirname}" || (download_untar_pkg "${url}" && mv "${name}" "${dirname}") && \
            cd "${dirname}" && \
            cat openloops.cfg.tmpl | sed 's/^#compile_extra.*/compile_extra = 1/' > openloops.cfg && \
            ./scons && \
            ./openloops libinstall -j"${NPROC}" ppzj ppzjj ppllj eett eehtt tbw eevvjj && \
            popd
}

build_pythia8 () {
    local url="${1}"
    local config="${2:-}"
    name="$(basename ${url} .tgz)"
    test -d "${name}" || download_untar_pkg "${url}"
    pushd "${name}" && \
        make_make_install "${config}" && \
        popd
}

build_rivet () {
    local url="${1}"
    local config="${2:-}"
    local name="$(basename ${url})"
    local version="$(basename $(dirname ${url}))" && \
        local dirname="rivet-${version}"
    test -f ${name} || \
        (wget --no-check-certificate --no-verbose "${url}" -O "${name}" && \
             chmod u+x "${name}" && \
             mkdir -p "${dirname}")
    pushd "${dirname}" && \
        # Wrap rivet-bootstrap and ENV variables into an eval statement.
        eval "${config} ../rivet-bootstrap" && \
        mv "${PREFIX}/rivetenv.csh" "${PREFIX}/bin" && \
        mv "${PREFIX}/rivetenv.sh" "${PREFIX}/bin" && \
        mv "${PREFIX}/yodaenv.sh" "${PREFIX}/bin" && \
        popd
}

# Download and build package with NAME from URL, log the progress to LOGDIR/NAME.log.
make_and_log () {
    local name="${1}"
    local url="${2}"
    local config="${3:-}"
    echo "Build ${name} ... " && \
        build_generic "${url}" "${config}" 1>"${LOG}/${name}.log" 2>&1
    if test $? -eq 0; then
        echo "... finished."
    else
        echo "... failed. See ${LOG}/${name}.log."
        exit 2
    fi
}

cmake_and_log () {
    local name="${1}"
    local url="${2}"
    local config="${3:-}"
    echo "Build ${name} ... " && \
        build_cmake_generic "${url}" "${config}" 1>"${LOG}/${name}.log" 2>&1
    if test $? -eq 0; then
        echo "... finished."
    else
        echo "... failed. See ${LOG}/${name}.log."
        exit 2
    fi
}

build_and_log_hepmc () {
    local name="${1}"
    local url="${2}"
    local config="${3:-}"
    echo "Build ${name} ... " && \
        build_hepmc2 "${url}" "${config}" 1>"${LOG}/${name}.log" 2>&1
    if test $? -eq 0; then
        echo "... finished."
    else
        echo "... failed. See ${LOG}/${name}.log."
        exit 2
    fi
}

build_and_log_ocaml () {
    local url="${1}"
    local name="OCaml"
    local config="${2:-}"
    echo "Build ${name} ... " && \
        build_ocaml "${url}" "${config}" 1>"${LOG}/${name}.log" 2>&1
    if test $? -eq 0; then
        echo "... finished."
    else
        echo "... failed. See ${LOG}/${name}.log."
        exit 2
    fi
}

build_and_log_openloops () {
    local url="${1}"
    local name="OpenLoops"
    local config="${2:-}"
    echo "Build ${name} ... " && \
        build_openloops "${url}" "${config}" 1>"${LOG}/${name}.log" 2>&1
    if test $? -eq 0; then
        echo "... finished."
    else
        echo "... failed. See ${LOG}/${name}.log."
        exit 2
    fi
}

build_and_log_pythia8 () {
    local url="${1}"
    local name="Pythia8"
    local config="${2:-}"
    echo "Build ${name} ... " && \
        build_pythia8 "${url}" "${config}" 1>"${LOG}/${name}.log" 2>&1
    if test $? -eq 0; then
        echo "... finished."
    else
        echo "... failed. See ${LOG}/${name}.log."
        exit 2
    fi
}

build_and_log_rivet () {
    local url="${1}"
    local name="Rivet"
    local config="${2:-}"
    echo "Build ${name} ... " && \
        build_rivet "${url}" "${config}" 1>"${LOG}/${name}.log" 2>&1
    if test $? -eq 0; then
        echo "... finished."
    else
        echo "... failed. See ${LOG}/${name}.log."
        exit 2
    fi
}

# #######################################################################


write_script () {
    name=$1
    date="$(date)"
    myself="$0"
    rm -f "${name}" && touch "${name}"
    cat <<EOF > "${name}"
#!/bin/bash
# #######################################################################
# WHIZARD build environment
# Date: ${date}
# Created by: ${myself}
# #######################################################################
export PATH=${PREFIX}/bin:\${PATH}
export LD_LIBRARY_PATH=${PREFIX}/lib:${PREFIX}/OpenLoops/lib:\${LD_LIBRARY_PATH}
# #######################################################################
EOF
    chmod 755 "${name}"
}

# #######################################################################

function help() {
    # sed -n 's/^##[[:space:]]//p' "${BASH_SOURCE[0]}" >&2
    sed -n 's/^##//p' "${BASH_SOURCE[0]}" >&2
    exit 1
}

# #######################################################################

HEPMC2_CONFIG=(
    -DCMAKE_INSTALL_PREFIX="${PREFIX}"
    -DCMAKE_CXX_FLAGS="--std=c++11"
    -Dmomentum:STRING=GEV
    -Dlength:STRING=MM
)

# Disable Python as the dependency is python-dev
LHAPDF_CONFIG=(
    CXXFLAGS="--std=c++11"
    --disable-python
)

FASTJET_CONFIG=(
    CXXFLAGS="--std=c++11"
    --enable-shared
    --disable-auto-ptr
    --enable-allcxxplugins
)

PYTHIA8_CONFIG=(
    --enable-debug
    --with-fastjet3="${PREFIX}"
    --with-lhapdf6="${PREFIX}"
    --with-hepmc2="${PREFIX}"
)

# PYTHIA8_DATA:PATH=PYTHIA8_DATA-NOTFOUND
# PYTHIA8_INCLUDE_DIR:PATH=PYTHIA8_INCLUDE_DIR-NOTFOUND
# PYTHIA8_LIBRARY:FILEPATH=PYTHIA8_LIBRARY-NOTFOUND
# PYTHIA8_hepmcinterface_LIBRARY:FILEPATH=PYTHIA8_hepmcinterface_LIBRARY-NOTFOUND
# PYTHIA8_lhapdfdummy_LIBRARY:FILEPATH=PYTHIA8_lhapdfdummy_LIBRARY-NOTFOUND

# CMAKE_INSTALL_PREFIX 	PATH 	Path where ROOT will be installed if make install is invoked or the “INSTALL” target is built.
# CMAKE_C_FLAGS 	STRING 	Extra flags to use when compiling C source files.
# CMAKE_CXX_FLAGS 	STRING 	Extra flags to use when compiling C++ source files.
# BUILD_SHARED_LIBS 	BOOL 	Flag indicating if shared libraries will be built. Its default value is ON.
# CMAKE_Fortran_COMPILER 	PATH 	Full path to the Fortran compiler. Alternatively you can specify the environment variable FC before invoking cmake
# CMAKE_C_COMPILER 	PATH 	Full path to the C compiler. Alternatively you can specify the environment variable CC before invoking cmake
# CMAKE_CXX_COMPILER 	PATH 	Full path to the C++ compiler. Alternatively you can specify the environment variable CXX before invoking cmake
# Use: FC, CC, CXX for Fortran, C and C++ Compiler

## * Root
##
##   We handle ROOT /just/ as a dependency of Delphes and HepMC3.
##   Therefore, we do not configure ROOT with any additional external tools, e.g. Pythia8.
##   We will only include mandatory dependencies, when needed.
##
## ** Note
##
##   Per default, we do not include the ROOT build.
##   However, the user *must* set an exisiting =ROOT_DIR= which points towards the top-level directory of an ROOT installation.
##   We /only/ check the existence of the top-level directory and assess that =ROOTConfig.cmake= exists.
##
## ** Example
##
##   ROOT_DIR=/usr/lib/cmake/ROOT/ROOTConfig.cmake
##
ROOT_CONFIG=(
    -DCMAKE_INSTALL_PREFIX="${PREFIX}"
)

# HepMC3 include ROOTIO

# Delphes comes with its own FastJet build chain.
DELPHES_CONFIG=(
    -DCMAKE_INSTALL_PREFIX="${PREFIX}"
    -DROOT_DIR="${ROOT_DIR}/cmake"
    -DPYTHIA8_INCLUDE_DIR="${PREFIX}/include"
    -DPYTHIA8_LIBRARY="${PREFIX}/lib/libpythia8.so"
)

# https://hepmc.web.cern.ch/hepmc/building.html
# Note: We drop the Python support (as we do not want to mess with any (custom or system) Python setup).
HEPMC3_CONFIG=(
    -DCMAKE_INSTALL_PREFIX="${PREFIX}"
    -DHEPMC3_ENABLE_PYTHON="OFF"
)

WHIZARD_CONFIG=(
    --prefix="${PREFIX}"
    --enable-lhapdf LHAPDF_DIR="${PREFIX}"
    --enable-hepmc HEPMC_DIR="${PREFIX}"
    --enable-pythia8 --with-pythia8="${PREFIX}"
    --enable-fastjet --with-fastjet="${PREFIX}"
    --enable-openloops --with-openloops="${PREFIX}/OpenLoops"
    --enable-python
    --disable-static
)

RIVET_CONFIG=(
    INSTALL_PREFIX="\"${PREFIX}\""
    MAKE="\"make -j${NPROC} -s V=0\""
    INSTALL_HEPMC=0
    INSTALL_FASTJET=0
    HEPMCPATH="\"${PREFIX}\""
    FASTJETPATH="\"${PREFIX}\""
)

## * Allowed modes
##
##   - autotools
##   - ocaml
##   - hepmc2
##   - lhapdf
##   - fastjet
##   - openloops
##   - root
##   - hepmc3
##   - delphes
##   - pythia8
##   - rivet
##   - whizard
##
##   Build chain: FastJet, HepMC2 and LHAPDF6 → Pythia8 → ROOT → Delphes, HepMC3
##
function main() {
    mkdir -p "${LOG}"
    mkdir -p "${BUILD}"
    cd "${BUILD}" || (echo "cd: failed to go to ${BUILD}" && exit 2)
    if test $# -eq 0 || test "$@" == ""; then
        local modes=(autotools ocaml hepmc2 hepmc3 lhapdf fastjet pythia8 openloops bin whizard)
    else
        local modes=("${@}")
    fi


    # Build chain: FastJet, HepMC and LHAPDF6 → Pythia8 → ROOT → Delphes, HepMC3
    for mode in "${modes[@]}"; do
        case $mode in
            autotools)
                (command -v make >/dev/null 2>&1     && test "${FORCE}" != "True") || make_and_log "make" "${url_make}"
                # https://lists.gnu.org/archive/html/bug-m4/2019-10/msg00000.html
                # We exclude m4 as long as there is no fix in sight.
                # (command -v m4 >/dev/null 2>&1       && test "${FORCE}" != "True") || make_and_log "m4" "${url_m4}"
                (command -v autoconf >/dev/null 2>&1 && test "${FORCE}" != "True") || make_and_log "autoconf" "${url_autoconf}"
                (command -v automake >/dev/null 2>&1 && test "${FORCE}" != "True") || make_and_log "automake" "${url_automake}"
                (command -v libtool >/dev/null 2>&1  && test "${FORCE}" != "True") || make_and_log "libtool" "${url_libtool}"
                ;;
            ocaml)
                build_and_log_ocaml "${url_ocaml}"
                ;;
            hepmc2)
                build_and_log_hepmc "HepMC2" "${url_hepmc}" "${HEPMC2_CONFIG[*]}"
                ;;
            lhapdf)
                make_and_log "LHAPDF6" "${url_lhapdf}" "${LHAPDF_CONFIG[*]}" && \
                    wget https://lhapdfsets.web.cern.ch/lhapdfsets/current/CT10.tar.gz -O- | tar xz -C "${PREFIX}"/share/LHAPDF && \
                    wget https://lhapdfsets.web.cern.ch/lhapdfsets/current/cteq6l1.tar.gz -O- | tar xz -C "${PREFIX}"/share/LHAPDF
                # lhapdf install cteq6 cteq6l1 cteq66 CT10
                ;;
            fastjet)
                make_and_log "FastJet" "${url_fastjet}" "${FASTJET_CONFIG[*]}"
                ;;
            openloops)
                build_and_log_openloops "${url_openloops}"
                ;;
            root)
                cmake_and_log "ROOT" "${url_root}" "${ROOT_CONFIG[*]}"
                ;;
            hepmc3)
                if test "${FORCE_ROOT}" = True; then
                    if ! test -d "${ROOT_DIR}/cmake" || ! test -f "${ROOT_DIR}/cmake/ROOTConfig.cmake"; then
                        cmake_and_log "ROOT" "${url_root}" "${ROOT_CONFIG[*]}"
                    fi
                    # export LD_LIBRARY_PATH=${LD_LIBRARY_PATH}:path_to_ROOT6_libraries
                    HEPMC3_CONFIG+=(
                        -DHEPMC3_ENABLE_ROOTIO="ON"
                        -DROOT_DIR="${ROOT_DIR}"
                    )
                else
                    HEPMC3_CONFIG+=(
                        -DHEPMC3_ENABLE_ROOTIO="OFF"
                    )
                fi
                cmake_and_log "HepMC3" "${url_hepmc3}" "${HEPMC3_CONFIG[*]}"
                ;;
            delphes)
                if test "${FORCE_ROOT}" = True; then
                    if ! test -d "${ROOT_DIR}/cmake" || ! test -f "${ROOT_DIR}/cmake/ROOTConfig.cmake"; then
                        cmake_and_log "ROOT" "${url_root}" "${ROOT_CONFIG[*]}"
                    fi
                fi
                cmake_and_log "Delphes" "${url_delphes}" "${DELPHES_CONFIG[*]}"
                ;;
            pythia8)
                build_and_log_pythia8 "${url_pythia8}" "${PYTHIA8_CONFIG[*]}"
                ;;
            rivet)
                build_and_log_rivet "${url_rivet}" "${RIVET_CONFIG[*]}"
                ;;
            whizard)
                make_and_log "WHIZARD" "${url_whizard}" "${WHIZARD_CONFIG[*]}"
                ;;
            whizard-config)
                echo "${WHIZARD_CONFIG[@]}"
                exit 0
                ;;
            bin)
                # Secure that the bin directory exists
                mkdir -p "${PREFIX}/bin"
                write_script "${PREFIX}/bin/whizard-build-env.sh"
                ;;
            help)
                help
                ;;
            *)
                echo "Unknown ${mode}... skipping"
                ;;
        esac
    done
}

FORCE_ROOT=False

## * Command-line options:
##
##   -h : print help message (or when none argument is presented)
##   -r : force build of ROOT
while getopts ":hr" opt; do
    case "${opt}" in
        h)
            help
            ;;
        r)
            FORCE_ROOT=True
            ;;
        \?)
            echo "Invalid option: -${OPTARG}" >&2
            exit 1
            ;;
        :)
            echo "Option -${OPTARG} requires an argument." >&2
            exit 1
            ;;
    esac
done
shift $((OPTIND - 1))

main "$@"

