#!/usr/bin/env python

import copy
import re
import yaml
import jinja2 as j2

from time import time
import sys
from os import getcwd, path, makedirs

templateLoader = j2.FileSystemLoader(searchpath=getcwd())
templateEnv = j2.Environment(loader=templateLoader)


class Process:
    name = ""
    directory = ""
    config = {
        'benchmark': False,
        'integrate': False,
        'results': True,
        'simulate': False,
        'n_events': 10000,
        'sample_format': 'hepmc',
        'rng_method': 'rng_stream',
        'seed': 0,
        'iterations': "",
        'integration_method': 'vamp2',
        'nlo_calculation': ''
    }

    def __init__(self, config, def_config={}):
        if "name" not in config:
            print("Process definition without an actual name!")
            exit()
        self.name = config.pop('name')
        self.config['directory'] = config.get("directory", "")
        if self.config['directory'] != "":
            if not path.exists(self.config['directory']):
                print("Create directory {}"
                      .format(self.config["directory"]))
                makedirs(self.config["directory"])
        print("|-- name {} in directory {}".format(
            self.name, self.config['directory']))
        # Apply defaults
        for key in def_config.keys():
            self.config[key] = def_config[key]
        for key in config.keys():
            self.config[key] = config.get(key, self.config[key])
        # Set seed to time modulo N
        if self.config['seed'] == 0:
            self.config['seed'] = int(time()) % 100000

    def __parse(self):
        sindarin = {}
        sindarin['name'] = self.name
        if self.config['integrate']:
            if self.config['iterations']:
                sindarin['iterations'] = "iterations = {}".format(
                    self.config['iterations'])
            sindarin['integrate'] = "integrate ({})".format(self.name)
        if self.config['simulate']:
            if 'checkpoint' not in self.config:
                self.config['checkpoint'] = int(self.config['n_events'] / 10)
            sindarin['checkpoint'] = "checkpoint = {}".format(
                self.config['checkpoint'])
            sindarin['n_events'] = "n_events = {}".format(
                self.config['n_events'])
            sindarin['sample_format'] = "sample_format = {}".format(
                self.config['sample_format'])
            sindarin['sample'] = "$sample = \"{}\"".format(
                self.name)
            sindarin['simulate'] = "simulate ({})".format(
                self.name)
        sindarin['rng_method'] = "$rng_method = \"{}\"".format(
            self.config['rng_method'])
        sindarin['integration_method'] = "$integration_method = \"{}\"".format(
                self.config['integration_method'])
        sindarin['seed'] = "seed = {}".format(
                self.config['seed'])
        if self.config['nlo_calculation']:
            sindarin['nlo_calculation'] = "{{ nlo_calculation = {} }}".format(
                ", ".join(self.config['nlo_calculation']))
        if self.config['results']:
            sindarin['results'] = 'printf "RES %8e %8e" (integral({}), error({}))'.format(self.name, self.name)
            return sindarin

    def save(self, template_filename):
        sindarin = self.__parse()
        name = self.name
        if self.config['integrate']:
            name += '-integrate'
        if self.config['simulate']:
            name += '-simulate'
        filename = "{}/{}.sin".format(self.config['directory'], name)
        template = templateEnv.get_template(template_filename)
        with open(filename, 'w') as outfile:
            outfile.write(template.render(sindarin))


def main():
    if (len(sys.argv) < 2):
        print("You have to specify at least on template!")
        exit(1)

    for file in sys.argv[1:]:
        filename = path.basename(file)
        process_id = re.sub('-template.sin', "", filename)
        # Use a common config.yml, else use "process_id-config.yml".
        if path.isfile("config.yml"):
            config_file = "config.yml"
        elif path.isfile(re.sub('-template.sin', "-config.yml")):
            config_file = re.sub('-template.sin', "-config.yml")
        else:
            print("You have to prepare a config.yml or {}-config.yml".format(process_id))
            exit

        print("Initialize template for {}".format(process_id))
        with open(config_file, 'r') as stream:
            for config in yaml.load_all(stream):
                # Find default config if set
                def_config = {}
                for instance in config[process_id]:
                    if 'default' in instance:
                        def_config = instance['default']
                        break
                for instance in config[process_id]:
                    if 'process' in instance:
                        if instance['process'].get('benchmark', False):
                            for i in list(range(0, 9)):
                                # Deep copy needed, because we change the object
                                process_instance = copy.deepcopy(instance['process'])
                                process_instance['directory'] += "/{}".format(2**i)
                                process = Process(process_instance, def_config)
                                process.save(file)
                        else:
                            process = Process(instance['process'], def_config)
                            process.save(file)


if __name__ == "__main__":
    main()
