#!/bin/bash
WORKDIR=`pwd`
WHIZARD_SCRIPTS="${HOME}/whizard-scripts"
OUTPUT="results"
LOG_DIR="log"

DEBUG="False"

function suffix () {
    local suffix=""
    if [ "${INTEGRATE}" = True ]; then
        suffix+="-integrate"
    fi
    if [ "${SIMULATE}" = True ]; then
        suffix="-simulate"
    fi
    echo ${suffix}
}

function result-to-table {
    local name=${1}

    if [ -f "${OUTPUT}/${name}.result" ]; then
        ${WHIZARD_SCRIPTS}/result-to-table.py ${OUTPUT}/${name}.result
    fi
}

RESULT_LOG="mpi.log.100"
function log-to-result () {
    local process=${1%/}

    local directory=${process}
    if [ -d "${directory}/128" ]; then
        directory+="/128"
    fi
    (set -x
     result=`cat ${directory}/${RESULT_LOG} | grep 'RES' | cut -d' ' -f2-`
     echo "${result}" > ${directory}/${process}.result
    )
}

function integrate-result {
    local directory=${1%/}
    local result=${2}

    mkdir -p "${OUTPUT}"
    echo "# Process Integral Error" > ${OUTPUT}/${result}.result
    for file in $(find ${directory} -type f -name "*.result" | grep -v "${OUTPUT}"); do
        if [ -s "${file}" ]; then
            (set -x
             echo "$(basename ${file} .result) $(cat ${file})" >> ${OUTPUT}/${result}.result
            )
        fi
    done
    cat "${OUTPUT}/${result}.result"
}

function average-run-time {
    local directory=${1%/}
    local process=$(basename ${directory})
    local subdirectory=$(dirname ${directory})

    mkdir -p "${OUTPUT}/${subdirectory}"
    ${WHIZARD_SCRIPTS}/average-run-time.py ${OUTPUT} ${directory}  2>&1 | tee ${process}-average-run-time.log
    if test ${DEBUG} = "True"; then
        echo "${OUTPUT}/${subdirectory}/${process}.time"
        cat "${OUTPUT}/${subdirectory}/${process}.time"
    fi
}

function run-time-speedup {
    local process=${1}

    (set -x
     ${WHIZARD_SCRIPTS}/run-time-speedup.py ${OUTPUT}/${process}.time
    ) 2>&1 | tee ${OUTPUT}/run-time-speedup.log
    cat "${OUTPUT}/${process}.speedup"
}

########################################################################
# Prepare WHIZARD run
########################################################################
RESULT_DIR="result"

## - init [process] [prefix]
##
##   Initializes process in directory prefix/process using the sindarin file process.sin.
##   When sindarin file is a template (-template.sin) the sindarin parser applies the settings given by the .yml-config.
##   SCRIPT: modify WHIZARD_SCRIPTS.
function init {
    local process=${1%.sin}
    local prefix=${2%/}
    local sinfile="${prefix:+${prefix}/}${process}.sin"
    local logfile="${LOG_DIR}/${prefix:+${prefix}.}${process}.log"

    if [ -f "${process}-template.sin" ]; then
        (set -x
         ${WHIZARD_SCRIPTS}/init-sindarin.py ${process}-template.sin
        ) 2>&1 | tee ${logfile}
    elif [ -f ${sinfile} ]; then
        (set -x
         mkdir -p ${prefix:+${prefix}/}${process} && cp ${sinfile} ${prefix:+${prefix}/}${process}
        ) 2>&1 | tee ${logfile}
    fi
}

## - collect [prefix=.] [lines=0]
##
##   collects dat-files (if exist) from given prefix directory and tests if file fulfills >= lines.
##   SCRIPT: modify RESULT_DIR="result"
function collect {
    local process=${1}
    local prefix=${2:-.}
    local lines=${3:-0}

    local directory="${prefix}/${process}"
    local datfile="${directory}/${process}.dat"
    if test -f ${datfile}; then
        linenumber=`cat ${datfile} | wc -l`
        if test "${DEBUG}" == "True"; then
            echo "$linenumber ($lines)"
        fi
        if (( ${linenumber} >= ${lines} )); then
            (set -x
             mkdir -p "${RESULT_DIR}/${directory}"
             cp -a ${datfile} "${RESULT_DIR}/${datfile}"
            )
        else
            echo "${process} ${prefix}/${process}"
        fi
    else
        echo "${process} ${prefix}/${process}"
    fi
}

########################################################################
# Run WHIZARD
########################################################################

RUN_TIME=True
MPIRUN=False
N_TASKS=""
HOSTFILE=""
WHIZARD_ENV_BIN="${HOME}/workspace/local/bin/whizard-build-env.sh"
WHIZARD_BIN="${HOME}/workspace/mpi/install/bin/whizard-setup.sh"
WHIZARD_REBUILD=()
WHIZARD_GLOBAL_OPTIONS=()

function profil () {
    local process=${1%.sin}
    local directory=${2:-"${process}"}
    local sinfile="${process}$(suffix).sin"

    local args=(
        ${WHIZARD_GLOBAL_OPTIONS[@]}
        ${WHIZARD_REBUILD[@]}

        ${sinfile}
    )

    shift ${#}                      # Do not remove...
    if [ -f "${WHIZARD_ENV_BIN}" ]; then
        source ${WHIZARD_ENV_BIN}
    fi
    if [ -f "${WHIZARD_BIN}" ]; then
        source ${WHIZARD_BIN}
    fi

    cmd="cd ${directory} &&"
    cmd+=" libtool --mode=execute perf record -g"
    cmd+=" whizard ${args[@]}"
    ( set -x
      eval "${cmd}"
    ) 2>&1 | tee ${process}-run-whizard.log
}

## - debug process [directory=${process}]
##
##   run WHIZARD in gdb under WHIZARD Build Env.
function debug () {
    local process=${1%.sin}
    local directory=${2:-"${process}"}
    local sinfile="${process}$(suffix).sin"
    local logfile="${LOG_DIR}/${process}-debug.log"

    local args=(
        ${WHIZARD_GLOBAL_OPTIONS[@]}
        ${WHIZARD_REBUILD[@]}

        ${sinfile}
    )

    shift ${#}                      # Do not remove...
    if [ -f "${WHIZARD_ENV_BIN}" ]; then
        source ${WHIZARD_ENV_BIN}
    fi
    if [ -f "${WHIZARD_BIN}" ]; then
        source ${WHIZARD_BIN}
    fi

    cmd="cd ${directory} &&"
    cmd+=" libtool --mode=execute gdb --args"
    cmd+=" whizard ${args[@]}"
    ( set -x
      eval "${cmd}"
    ) 2>&1 | tee "${logfile}"
}

## - run process [directory=${process}] [np=3]
##
##   run WHIZARD under WHIZARD Build Env, optionally with MPI with np processses, hostfile and logging.
##   Additionally, the runtime is taken.
function run () {
    local process=${1%.sin}
    local directory=${2:-"${process}"}
    local sinfile="${process}$(suffix).sin"
    local np=${3:-"${N_TASKS}"}
    local logfile="${LOG_DIR}/${process}-run.log"

    local args=(
        ${WHIZARD_GLOBAL_OPTIONS[@]}
        ${WHIZARD_REBUILD[@]}

        ${sinfile}
    )

    shift ${#}                      # Do not remove...
    if [ -f "${WHIZARD_ENV_BIN}" ]; then
        source ${WHIZARD_ENV_BIN}
    fi
    if [ -f "${WHIZARD_BIN}" ]; then
        source ${WHIZARD_BIN}
    fi

    cmd="cd ${directory} &&"
    if [ "$RUN_TIME" = True ]; then
        cmd+=" /usr/bin/time --append -f %e -o ${process}$(suffix).time"
    fi
    if [ "$MPIRUN" = True ]; then
        cmd+=" mpirun"
        if [ "${np}" ]; then
            cmd+=" -np ${np}"
        fi
        if [ "$HOSTFILE" = True ]; then
            cmd+=" --hostfile ${HOSTFILE}"
        fi
        if [ $MPILOG ]; then
            cmd+=" --output-filename ${MPILOG}"
        fi
    fi
    cmd+=" whizard ${args[@]}"
    ( set -x
      eval "${cmd}"
    ) 2>&1 | tee "${logfile}"
}

########################################################################
# Analysis and Plotting
########################################################################

function analyze () {
    local process=${1}
    local directory=${2:-"."}
    (set -x
     cd ${directory} && \
         rivet --quiet \
               -H ${ANALYSIS}.yoda \
               --analysis-path=${WORKDIR} \
               --analysis=${ANALYSIS} *.hepmc
    ) 2>&1 | tee ${process}-anaylze.log
}

WHIZARD_SCRIPT_PATH="${HOME}/whizard-scripts/whizard.sh"

function help() {
    sed -n 's/^##//p' ${WHIZARD_SCRIPT_PATH}
    echo "Multi-Purpose script for running or timing WHIZARD and analysing."
    echo "Usage: [function] [options]"
    echo "========================================================================"
    echo "prepare-sindarin [sin-file] [vamp2 [true]]"
    echo "	The filename without .sin is the process."
    echo "run [process] [directory=process]"
    echo "average-run-time [directory]"
    echo "	Calculate the average run-time in all *.time files and save to OUTPUT as [directory].result."
    echo "run-time-speedup [process]"
    echo "========================================================================"
}

########################################################################
# Validation
########################################################################

function compare () {
    local process=${1}
    local mpi=${MPIRUN}
    for method in vamp vamp2; do
        if [ "$method" = vamp ]; then
            MPIRUN=False
        else
            MPIRUN=${mpi}
        fi
        http://www.hepforge.org/archive/lhapdf/LHAPDF-6.2.0.tar.gz       run "${process}" "${method}" && \
            anaylze "${process}" "${method}"
    done

    rivet-mkhtml --mc-errs --pdf --cm --single \
                 -c ${ANALYSIS}.plot \
                 vamp/${ANALYSIS}.yoda:"Title=VAMP:ErrorBands=1:ErrorBandColor=blue:ErrorBandOpacity=0.5:LineColor=blue" \
                 vamp2/${ANALYSIS}.yoda:"Title=VAMP2:ErrorBands=1:ErrorBandColor=red:ErrorBandOpacity=0.5LineColor=red"
}

INTEGRATE=False
SIMULATE=False
function integrate () {
    INTEGRATE=True
    SIMULATE=False
    run-time "$@"
}

function simulate () {
    INTEGRATE=True
    SIMULATE=True
    run-time "$@" && \
        anaylze "$@"
}

SUFFIX=""
function main() {
    local run="$1"
    shift

    if test ! -d "${LOG_DIR}"; then
        mkdir -p "${LOG_DIR}"
    fi
    "${run}" "$@"
}
